dir = File.dirname(__FILE__)
Sketchup.require File.join(dir, 'core.rb')

# Sketchup nested namespace.
# This module is the core to gather and store SketchUp hWnds and IDs.
# As well, it provides functions that has to do with the whole Sketchup thread.
module AMS::Sketchup

  # Versions prior to 2.0.0 were not supported with the proper namespace and the version constant.
  VERSION = '2.0.1' # May 06, 2013


  # MODULE VARIABLES
  @@process_id   = nil # process id
  @@thread_id    = nil # thread id
  @@main_window  = nil # hWnd to SketchUp's top, parent window
  @@control_bars = []  # [top, bottom, left, right] - toolbar containers
  @@status_bars  = []  # [status bar, scenes bar]
  @@viewport     = nil # hWnd to view window
  @@menu_bar     = nil # hWnd to menu bar
  #~ @@window       = nil # temp variable
  @@thread_toolbars   = []
  @@thread_dialogs    = []
  @@dialogs_to_ignore = []


  # Create one callback to find hWnd to the main window.
  @@find_main_hwnd ||= AMS::Win32::API::Callback.new('IP', 'I'){ |hwnd, lParam|
    # Get thread and process id of current enumerated window hWnd.
    p = 0.chr*4
    threadID = AMS::GetWindowThreadProcessId.call(hwnd, p)
    pid = p.unpack('l')[0]
    # Skip the rest operation if pid of current enumerated window does not match current pid.
    next true if (pid != @@process_id)
    # Check for the match in class name
    cname = 0.chr*50
    AMS::GetClassName.call(hwnd, cname, cname.size)
    cname.strip! # Removes leading and trailing whitespace
    # Search for match; SketchUp class name is not constant, only some parts are same.
    if ([cname[0,3], cname[13,1], cname[24,8]] == ['Afx', 'b', '00000006'])
      @@thread_id = threadID
      @@main_window = hwnd
      # Break out of the block/loop if found
      next false
    end
    # Otherwise, continue enumeration
    true
  }

  # All SU child windows will be enumerated through this callback function to locate
  # hWnds to status bar, scenes bar, toolbar containers, and viewport window.
  @@find_subwindows ||= AMS::Win32::API::Callback.new('IP', 'I'){ |hwnd, lParam|
    cname = 0.chr*50
    AMS::GetClassName.call(hwnd, cname, cname.size)
    cname.strip! # Removes leading and trailing whitespace
    # Collect hWnds to all four toolbar container
    @@control_bars.push(hwnd) if (cname == 'AfxControlBar80u')
    # Collect hWnds to status bar and scenes bar
    @@status_bars.push(hwnd) if (cname == 'msctls_statusbar32')
    # Get viewport hWnd
    @@viewport = hwnd if (cname == 'AfxFrameOrView80u')
    true # continue enumeration
  }

  # This callback will be used to find the hwnd of the specified window's caption text.
  #~ @@get_wind_by_caption ||= AMS::Win32::API::Callback.new('IS', 'I'){ |hwnd, lParam|
  #~   size = AMS::GetWindowTextLength.call(hwnd)
  #~   lpStr = 0.chr*size
  #~   AMS::GetWindowText.call(hwnd, lpStr, (lpStr.size+1))
  #~   if (lpStr == lParam)
  #~     @@window = hwnd
  #~     next false
  #~   end
  #~   true
  #~ }

  # This callback function will be used to get dialogs or floating toolbars.
  @@enum_windows ||= AMS::Win32::API::Callback.new('IP', 'I'){ |hwnd, lParam|
    if (lParam == "find thread toolbars")
      cname = 0.chr*50
      AMS::GetClassName.call(hwnd, cname, cname.size)
      cname = cname.strip
      cmatch = ((cname[0,3] == "Afx") and (cname[13,1] == "8") and (cname[24,8] == "00000000"))
      @@thread_toolbars.push(hwnd) if (cmatch and (AMS::IsWindowVisible.call(hwnd) != 0) and (@@thread_toolbars.include?(hwnd) == false))
      next true
    end
    if (lParam == "find thread dialogs")
      cname = 0.chr*50
      AMS::GetClassName.call(hwnd, cname, cname.size)
      cname = cname.strip
      match = ((cname == "#32770") and (AMS::IsWindowVisible.call(hwnd) != 0) and (@@thread_dialogs.include?(hwnd) == false))
      @@thread_dialogs.push(hwnd) if match
      next true
    end
    false
  }


  class << self  #  <--<< THIS module's anonymous singleton proxy class.

    # Retrieves process id
    def process_id(); @@process_id end

    alias pid process_id


    # Retrieves thread id
    def thread_id(); @@thread_id end

    alias tid thread_id


    # Retrieves hWnd to SketchUp's top, parent window
    def main_window(); @@main_window end

    alias parent_window main_window


    # Retrieves hWnd to original menubar
    def menu_bar(); @@menu_bar end


    # Retrieves an array of hWnds to all four toolbar containers.
    # Args::
    #   Void
    # Returns::
    #   Array: [top, bottom, left, right] - toolbar containers
    #
    def toolbar_containers(); @@control_bars end

    alias control_bars toolbar_containers


    # Retrieves hWnd to status bar
    def status_bar(); @@status_bars[0] end


    # Retrieves hWnd to scenes bar
    def scenes_bar(); @@status_bars[1] end


    # Retrieves hWnd to viewport window
    def viewport(); @@viewport end


    # Attempts to find hWnd to SketchUp's top, parent window
    # Args::
    #   Integer: Process ID. Default is current process ID.
    # Returns::
    #   Bool: true - found, false - not found
    #
    def findMainWindow(pid = Object::Process.pid)
      @@process_id = pid
      @@main_window = nil
      AMS::EnumWindows.call(@@find_main_hwnd, nil)
      (@@main_window != nil)
    end


    # Attempts to gather hWnds to the following child windows of SU main window:
    #   - four toolbar containers
    #   - status bar
    #   - scenes bar
    #   - and viewport
    # Args::
    #   Void
    # Returns::
    #   Bool: true - successfully gathered hWnds, false - hWnd to main window is undefined.
    #
    def collectSubWindows()
      # Return false if hWnd to main window is undefined.
      return false unless @@main_window
      @@menu_bar = AMS::GetMenu.call(@@main_window)
      AMS::EnumChildWindows.call(@@main_window, @@find_subwindows, nil)
      true
    end


    # Searches throught SU thread windows to find the window with the specified caption text.
    # Args::
    #   text: string caption text
    # Returns::
    #   windows's hWnd, or Nil - if not found
    #
    def getWindowByCaption(text)
      text = text.to_s
      #~ @@window = nil
      #~ AMS::EnumThreadWindows.call(@@thread_id, @@get_wind_by_caption, text)
      #~ @@window
      wind = AMS::FindWindow.call(nil, text)
      (wind == 0) ? nil : wind
    end


    # This method is used to show/hide floating toolbars,
    # the toolbars that are not embedded into SU toolbar containers.
    def showFloatingToolbars(state = true)
      state = [true,1].include?(state)
      # Show
      if state
        @@thread_toolbars.each{ |hwnd| AMS::ShowWindow.call(hwnd, 5) }
        @@thread_toolbars.clear
        return true
      end
      # Hide
      AMS::EnumThreadWindows.call(@@thread_id, @@enum_windows, "find thread toolbars")
      temp = @@thread_toolbars.dup
      temp.each{ |hwnd|
        if (AMS::IsWindow.call(hwnd) == 0)
          @@thread_toolbars.delete(hwnd)
          next
        end
        AMS::ShowWindow.call(hwnd, 0)
      }
      temp = nil
      true
    end


    # This method is used to show/hide dialogs, including the Ruby Console.
    def showDialogs(state = true)
      state = [true,1].include?(state)
      # Update the @@dialogs_to_ignore queue
      @@dialogs_to_ignore.each{ |hwnd|
        @@dialogs_to_ignore.delete(hwnd) if (AMS::IsWindow.call(hwnd) == 0)
      }
      # Show
      if state
        @@thread_dialogs.each{ |hwnd|
          next if @@dialogs_to_ignore.include?(hwnd)
          AMS::ShowWindow.call(hwnd, 1) # window gets corrupted
        }
        AMS::Sketchup::MainWindow.refresh # fixes corruption
        @@thread_dialogs.clear
        return true
      end
      # Hide
      AMS::EnumThreadWindows.call(@@thread_id, @@enum_windows, "find thread dialogs")
      temp = @@thread_dialogs.dup
      temp.each{ |hwnd|
        if (AMS::IsWindow.call(hwnd) == 0)
          @@thread_dialogs.delete(hwnd)
          next
        end
        next if @@dialogs_to_ignore.include?(hwnd)
        AMS::ShowWindow.call(hwnd, 0) # window gets corrupted
      }
      temp = nil
      true
    end


    # This method is used to elude dialog from the <tt>showDialogs</tt> and <tt>closeDialogs</tt> operation.
    # Args::
    #   hwnd: a window hWnd
    # Returns::
    #   status: true - dialog ignored, false - the dialog is already ignored or the given value is not a dialog
    #
    def ignoreDialog(hwnd)
      hwnd = hwnd.to_i
      return false if @@dialogs_to_ignore.include?(hwnd)
      # Verify given hwnd is a window
      return false if (AMS::IsWindow.call(hwnd) == 0)
      # Verify given hwnd is a dialog
      cname = 0.chr*50
      AMS::GetClassName.call(hwnd, cname, cname.size)
      cname = cname.strip
      return false if (cname != "#32770")
      # Verify given hwnd belongs to current thread
      id = AMS::GetWindowThreadProcessId.call(hwnd, nil)
      return false if (id != @@thread_id)
      # Push into queue once the hwnd met all the demands
      @@dialogs_to_ignore.push(hwnd)
      true
    end


    # This method is used to include dialog in the <tt>showDialogs</tt> operation.
    # By default all dialogs are included.
    # This method is used to remove the dialog from the ignore list.
    # Args::
    #   hwnd: a window hWnd
    # Returns::
    #   status: true - dialog included, false - the dialog is already included or the given value is not a dialog
    #
    def includeDialog(hwnd)
      return false unless @@dialogs_to_ignore.include?(hwnd)
      @@dialogs_to_ignore.delete(hwnd)
      true
    end


    # This method is used to close all the visible floating toolbars, and
    # all the floating toolbars that were hidden by the <tt>showFloatingToolbars</tt> function.
    def closeFloatingToolbars
      AMS::EnumThreadWindows.call(@@thread_id, @@enum_windows, "find thread toolbars")
      @@thread_toolbars.each{ |hwnd|
        next if (AMS::IsWindow.call(hwnd) == 0)
        AMS::SendMessage.call(hwnd, 0x0010, nil, nil) # WM_CLOSE
      }
      @@thread_toolbars.clear
      true
    end


    # This method is used to close all the visible dialogs, and
    # all the dialogs that were hidden by the <tt>showDialogs</tt> function.
    def closeDialogs
      AMS::EnumThreadWindows.call(@@thread_id, @@enum_windows, "find thread dialogs")
      temp = @@thread_dialogs.dup
      temp.each{ |hwnd|
        next if @@dialogs_to_ignore.include?(hwnd)
        if (AMS::IsWindow.call(hwnd) != 0)
          AMS::SendMessage.call(hwnd, 0x0010, nil, nil) # WM_CLOSE
        end
        @@thread_dialogs.delete(hwnd)
      }
      temp = nil
      #@@thread_dialogs.clear
      #@@dialogs_to_ignore.clear
      true
    end


    # Sent as a signal that a window or an application should terminate.
    # This function does the same operation as pressing the close button.
    def close
      # Use WM_CLOSE (0x0010) message: http://msdn.microsoft.com/en-us/library/windows/desktop/ms632617(v=vs.85).aspx
      # Note! It's important to call this message via PostMeessage or via SendMessage;
      # otherwise, using DefWindowProc will close the window improperly.
      AMS::SendMessage.call(hwnd, 0x0010, nil, nil)
    end


    # This function forces SU to close.
    def endTask
      AMS::EndTask.call(hwnd, false, true)
    end

  end # proxy class


  class MyAppObserver < Sketchup::AppObserver
    def onQuit
      # Usually the hidden statusbar would remain hidden on next launch; it does not show-up by default.
      # Unhide it on close.
      AMS::Sketchup::StatusBar.show
    end
  end # class MyAppObserver


  unless file_loaded?(__FILE__)
    Object::Sketchup.add_observer(MyAppObserver.new)
    file_loaded(__FILE__)
    # Find windows and load all files in the sketchup folder
    status = self.findMainWindow
    self.collectSubWindows
    dir = File.dirname(__FILE__)
    path =  File.join(dir, 'sketchup')
    files = Dir.glob(File.join(path, '*.{rb, rbs}'))
    files.each{ |file| Sketchup.require file }
  end

end # module AMS::Sketchup
