#Sketchup.require 'ams_WindowSettings/sketchup window.rb'

# Main window nested namespace.
# Main window is SketchUp's top, parent window.
module AMS::Sketchup::MainWindow

  # MODULE VARIABLES
  @@exstyle   = 0
  @@style     = 0
  @@placement = 0
  @@fs_state  = false # full screen state
  @@delay     = 0.4 # change delay
  @@last_time = Time.new # the last time SU was set full screen

  hDC = AMS::GetWindowDC.call(nil)
  cx = AMS::GetDeviceCaps.call(hDC, 8) # 8 - HorzRes
  cy = AMS::GetDeviceCaps.call(hDC, 10) # 10 - VertRes
  AMS::ReleaseDC.call(AMS::Sketchup.main_window, hDC)
  @@res = [cx,cy]

  class << self  #  <--<< THIS module's anonymous singleton proxy class.

    # Retrieves hWnd to SketchUp's top, parent window
    def hwnd
      AMS::Sketchup.main_window
    end

    alias handle hwnd


    # Retrieves window's class name
    def class_name
      cname = 0.chr*50
      AMS::GetClassName.call(hwnd, cname, cname.size)
      cname.strip
    end


    # Retrieves windows's title text
    def caption
      size = AMS::GetWindowTextLength.call(hwnd)
      lpStr = 0.chr*size
      AMS::GetWindowText.call(hwnd, lpStr, (lpStr.size+1))
      lpStr
    end


    # This function is used to set window's title text
    def caption=(text)
      text = text.to_s
      return false if (text == caption)
      AMS::SetWindowText.call(hwnd, text)
      AMS::Sketchup::SettingsObserver.onCaptionChanged(text)
      true
    end


    # This function is used to get the dimensions of the bounding rectangle of the window.
    # The dimensions are given in screen coordinates that are relative to the upper-left corner of the screen.
    # It retrieves the screen coordinates of the upper-left and lower-right corners of the window.
    # Returns an array: [x1,y1, x2,y2]
    def rect
      lpRect = 0.chr*16
      AMS::GetWindowRect.call(hwnd, lpRect)
      lpRect.unpack('l*')
    end


    # Retrieves the screen coordinates of the upper-left corner of the window
    # Returns an array: [x,y]
    def origin; rect[0..1] end


    # Retrieves window's size in pixels
    # Returns an array: [width,height]
    def size
      r = rect
      w = r[2] - r[0]
      h = r[3] - r[1]
      [w, h]
    end


    # Retrieves window's width
    def width; size[0] end


    # Retrieves window's height
    def height; size[1] end


    # Returns true/false, specifying whether the window has keyboard focus and is on top
    def isActive?
      s1 = (AMS::GetFocus.call() == hwnd)
      s2 = (AMS::GetActiveWindow.call == hwnd)
      s3 = (AMS::GetForegroundWindow.call == hwnd)
      state = (s1 and s2 and s3)
      return state
    end


    # Returns true/false, specifying whether the window is minimized
    def isIconic?; (AMS::IsIconic.call(hwnd) != 0) end

    alias isMinimized? isIconic?


    # Returns true/false, specifying whether the window is maximized
    def isZoomed?; (AMS::IsZoomed.call(hwnd) != 0) end

    alias isMaximized? isZoomed?


    # Returns true/false, specifying whether the window is restored
    def isRestored?; !(isIconic? or isZoomed?) end


    # Returns true/false, specifying whether SketchUp is full screen
    def isFullScreen?; @@fs_state end


    # Brings window to the top and sets keyboard focus
    def setActive
      AMS::BringWindowToTop.call(hwnd)
      AMS::SetForegroundWindow.call(hwnd)
      AMS::SetActiveWindow.call(hwnd)
      AMS::SetFocus.call(hwnd)
      true
    end


    # This function is used to minimize the window.
    def minimize; AMS::ShowWindow.call(hwnd, 2) end


    # This function is used to maximize the window.
    def maximize; AMS::ShowWindow.call(hwnd, 3) end


    # This function is used to restore the window.
    def restore
      setFullScreen(false) # Make sure the window maintains its original styles
      AMS::ShowWindow.call(hwnd, 9)
    end


    # This function is used to set SketchUp full screen.
    # Args::
    #   Bool: true - set fullscreen, false - set original settings
    # Returns::
    #   Bool: true - success, false - state not changed
    #
    def setFullScreen(state = true)
      state = [true,1].include?(state)
      return false if (@@fs_state == state) or (Time.now < (@@last_time + @@delay))
      if state
        # Save original styles and placement
        @@exstyle = AMS::GetWindowLong.call(hwnd, -20)
        @@style = AMS::GetWindowLong.call(hwnd, -16)
        @@placement = 0.chr*44
        AMS::GetWindowPlacement.call(hwnd, @@placement)
        # Set full screen styles and placement
        exstyle = 0
        to_remove  = 0x10CC0000 # ( WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_VISIBLE )
        to_include = 0x01000000 # ( WS_MAXIMIZE )
        style = (@@style & ~ to_remove) | to_include
        flags = 0x0064 # ( SWP_SHOWWINDOW | SWP_NOZORDER | SWP_FRAMECHANGED )
		AMS::Sketchup::Viewport.repaint
      else
        # Set original styles and placement
        exstyle = @@exstyle
        style = @@style
        flags = 0x0067 # ( SWP_SHOWWINDOW | SWP_NOZORDER | SWP_FRAMECHANGED | SWP_NOSIZE | SWP_NOMOVE )
        AMS::SetWindowPlacement.call(hwnd, @@placement)
      end
      # Set settings
      AMS::LockWindowUpdate.call(hwnd)
      AMS::SetWindowLong.call(hwnd, -20, exstyle)
      AMS::SetWindowLong.call(hwnd, -16, style)
      AMS::LockWindowUpdate.call(nil)
      AMS::SetWindowPos.call(hwnd, 0, 0, 0, @@res[0], @@res[1], flags) # resposition window
      @@fs_state = state
      @@last_time = Time.now
      AMS::Sketchup::SettingsObserver.onSetFullScreen(state)
      true
    end


    # This function is used to move and size the window.
    def move(x = 240, y = 212, width = 800, height = 600)
      restore
      AMS::MoveWindow.call(hwnd, x, y, width, height, true)
    end


    # The 'refresh' function repositions all child windows of the main window.
    # It must be called after hidding certain elements of the window, such as status bar.
    # NOTE: In the current version, this funtion directly calls WM_SIZE message,
    # which is not the most efficient way to refresh the window.
    # Improvements to this function are attempted!
    def refresh
      lpRect = 0.chr*16
      AMS::GetWindowRect.call(hwnd, lpRect)
      rect = lpRect.unpack('l*')
      x = rect[0]
      y = rect[1]
      cx = rect[2]-rect[0]
      cy = rect[3]-rect[1]
      # From obervations, it's probably not required to call WM_SIZING.
      # WM_SIZING (0x0214): http://msdn.microsoft.com/en-us/library/windows/desktop/ms632647(v=vs.85).aspx
      # AMS::SendMessage.call(hwnd, 0x0214, 1, lpRect) # WM_SIZING (0x0214)
      wParam = 0 # SIZE_RESTORED
      wParam = 1 if isIconic? # SIZE_MINIMIZED
      wParam = 2 if isZoomed? # SIZE_MAXIMIZED
      # WM_SIZE (0x0005): http://msdn.microsoft.com/en-us/library/windows/desktop/ms632646(v=vs.85).aspx
      # See here how to set high-order word and low-order word: http://www.quickmacros.com/help/Other/IDP_HIWORD.html
	  #AMS::Sketchup::Viewport.repaint
      AMS::PostMessage.call(hwnd, 0x0005, wParam, cy<<16|cx) # WM_SIZE (0x0005)
      true
    end

  end # proxy class

end # module AMS::Sketchup::MainWindow
